/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigator.h"
#include <iterator>
#include <Eigen/src/Geometry/Transform.h>

#include "ArmarXCore/core/services/tasks/TaskUtil.h"

#include "RobotAPI/libraries/core/remoterobot/RemoteRobot.h"

// Navigation
#include "Navigation/libraries/core/types.h"
#include "Navigation/libraries/util/util.h"
#include <Navigation/libraries/factories/NavigationStackFactory.h>
#include <Navigation/libraries/server/Navigator.h>

namespace armarx::nav::components
{

    Navigator::Navigator() : executor{platformUnit}
    {
    }

    Navigator::~Navigator()
    {
    }

    void Navigator::onInitComponent()
    {
    }

    void Navigator::onConnectComponent()
    {
    }

    void Navigator::onDisconnectComponent()
    {
    }

    void Navigator::onExitComponent()
    {
    }

    void Navigator::updateContext()
    {
        scene.staticScene = staticScene();
        scene.robot = getRobot();
    }

    std::string Navigator::getDefaultName() const
    {
        return "Navigator";
    }

    std::vector<core::Pose> convert(const std::vector<Eigen::Matrix4f>& wps)
    {
        std::vector<core::Pose> p;
        p.reserve(wps.size());
        std::transform(wps.begin(), wps.end(), std::back_inserter(p), [](const auto & p)
        {
            return Eigen::Affine3f(p);
        });
        return p;
    }

    void Navigator::moveTo(const std::vector<Eigen::Matrix4f>& waypoints,
                           const aron::data::AronDictPtr& stackConfig,
                           const std::string& navigationMode,
                           const Ice::Current&)
    {
        updateContext();

        server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, scene);
        
        server::Navigator navigator{stack, scene, executor};

        navigator.moveTo(convert(waypoints), core::NavigationFramesMap.from_name(navigationMode));
    }

    void Navigator::moveTowards(const Eigen::Vector3f& direction,
                                const aron::data::AronDictPtr& stackConfig,
                                const std::string& navigationMode,
                                const Ice::Current&)
    {
        server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, scene);
        server::Navigator navigator{stack, scene, executor};

        navigator.moveTowards(direction, core::NavigationFramesMap.from_name(navigationMode));
    }

    void Navigator::stop(const Ice::Current&)
    {
        //server::Navigator navigator;
        //navigator.stop();
    }

    armarx::PropertyDefinitionsPtr Navigator::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        def->component(platformUnit);

        // Add a required property.
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.
        // def->optional(properties.numBoxes, "p.box.Number", "Number of boxes to draw in ArViz.");

        return def;
    }

    core::StaticScene Navigator::staticScene()
    {
        core::StaticScene scene;

        objpose::ObjectPoseSeq objectPoses = ObjectPoseClientPluginUser::getObjectPoses();

        scene.objects = util::asSceneObjects(objectPoses);

        return scene;
    }

    VirtualRobot::RobotPtr Navigator::getRobot()
    {
        auto robot = RemoteRobot::createLocalCloneFromFile(
                         getRobotStateComponent(), VirtualRobot::RobotIO::RobotDescription::eFull);
        // auto robot = RemoteRobot::createLocalClone(getRobotStateComponent());
        ARMARX_CHECK_NOT_NULL(robot);

        // ARMARX_INFO << "Resetting robot";
        // robot = VirtualRobot::RobotIO::loadRobot(robot->getFilename());

        RemoteRobot::synchronizeLocalClone(robot, getRobotStateComponent());

        return robot;
    }

} // namespace armarx::nav::components
